@@ -261,6 +261,10 @@ class Field:
         result.validators = self.validators[:]
         return result
 
+    def _clean_bound_field(self, bf):
+        value = bf.initial if self.disabled else bf.data
+        return self.clean(value)
+
 
 class CharField(Field):
     def __init__(
@@ -694,6 +698,10 @@ class FileField(Field):
     def has_changed(self, initial, data):
         return not self.disabled and data is not None
 
+    def _clean_bound_field(self, bf):
+        value = bf.initial if self.disabled else bf.data
+        return self.clean(value, bf.initial)
+
 
 class ImageField(FileField):
     default_validators = [validators.validate_image_file_extension]
@@ -1294,7 +1302,7 @@ class GenericIPAddressField(CharField):
         self.unpack_ipv4 = unpack_ipv4
         self.default_validators = validators.ip_address_validators(
             protocol, unpack_ipv4
-        )[0]
+        )
         super().__init__(**kwargs)
 
     def to_python(self, value):
